#-------------------------- clear all variables and objects ------------------------------------
rm(list = ls())

# Mismatch project aims to see if biases in perceived attitudes can explain the anti-conformism behavior
# The autors: Guillaume Deffuant and Omid Roozmand
# Universit? Clermont Auvergne, INRAE, LISC, Aubi?re, France
# Email: guillaume.deffuant@irstea.fr
# -------------------------------------------------------------------------------------------------------

# For calculating correlations, regressions and plotting, we need the following packages. They need to be installed first
#install.packages("corrplot")
#install.packages("hexbin")
#install.packages("Hmisc")

#loading libraries
library(corrplot)
#install.packages("lhs")
library(hexbin)
library(Hmisc)

#main function (mismatch): this function does the main following tasks:
# - 
# - each agent selects its subjective and perceived norm network.
# - each agent computes the value of its subjective and perceived norm.
# - 

# and returns the mean of intention of agents in the following four categories:
# 1: low attitude, low perceived norm
# 2: low attitude, high perceived norm
# 3: high attitude, low perceived norm
# 4: high attitude, high perceived norm
#
# Also it returns the correlation coefficients and pvalues of attitudes, sbjective norm, 
# intention, and perceived norm of the following two categories:
# 1: low attitude agents
# 2: high attitude agents
#
# It also returns the regression between intention and perceived norm f the two following categories:
# 1: low attitude agents
# 2: high attitude agents
mismatch = function(nloops, nagents, n1, n2, methodSN, allrest, N1N2Method, alpha, coef, rho, beta, tauM, sigma, delta)
{
  atts = createAttitudes(nagents, sigma, delta)
  omega = 1
  aver = 0
  runCanBeUsed = 1
  atts[-1]
  
  nbs = 1:nagents
  sns = NULL
  ints = NULL
  pns = NULL
  netAll = seq(from=1, to=nagents, by=1)
  for (i in nbs) {
    myAttitude = atts[i]
    
    # ------------------------------- building n1n2 network: combination of both subjective and perceived norm network
    net12 = NULL
    
    # selects the whole network randomly. There are also some other possibilities to select the network. However, for the tests we use random selection for the first population
    if (N1N2Method == "random"){
      net12 = sample(x = nbs, size = n1+n2, replace = FALSE, prob = NULL)      
    }
    
    # ------------------------------- building subjective norm (sn) network based on probbaility ------   
    net1 = NULL
    if(methodSN == "snP"){
      net1 = sn_prob(net12, alpha, myAttitude, n1, atts)
    }
    
    
    # ------------------------------- computing threshold, subjective norm, and intention
    thh = th(tauM, omega, atts[i], aver, beta)
    sn =  pn_basedOn_threshold(rho,atts, thh, atts[i], net1)
    sns = c(sns, sn)
    ints = c(ints, (atts[i]+sn)/2)
    
    # ------------------------------- building perceived norm (pn) network --- computing perceived norm        
    net2 = NULL
    if (allrest == "rest") 
    {
      net2 = pn_rest(net12, net1)            
      pn = pn_basedOn_threshold(rho,atts, thh, atts[i], net2)      
    }
    else if (allrest == "all")
    {
      pn = pn_basedOn_threshold(rho,atts, thh, atts[i], net12)
    }
    else 
      return("the value should be: all or rest")
    
    pns = c(pns, pn)
  }
  
  #-- mean and standard deviation of attitudes
  ma = mean(atts)
  sda = sd(atts)
  #-- mean and standard deviation of perceived norm
  mpn = mean(pns)
  sdpn = sd(pns)
  
  #-- low/high thresholds for attitude and perceived norm
  lathres = ma - (sda*coef)
  hathres = ma + (sda*coef)
  lpnthres = mpn - (sdpn*coef)
  hpnthres = mpn + (sdpn*coef)
  
  #-- four categories to compute intention
  lalpn = NULL # low attotide low perceived group norm
  lahpn = NULL # low attotide high perceived group norm
  halpn = NULL # high attotide low perceived group norm
  hahpn = NULL # high attotide high perceived group norm
  
  #-- low/high attitude categories
  la = NULL
  ha = NULL
  
  #-- creating 6 categories: la, ha, lalpn,lahpn, halpn, hahpn
  for (i in nbs){
    if (atts[i] < lathres){
      la = c(la, i)
      if(pns[i]<lpnthres) lalpn = c(lalpn, i)
      if(pns[i]>hpnthres) lahpn = c(lahpn, i)
    }
    if (atts[i] > hathres){
      ha = c(ha , i)
      if(pns[i]<lpnthres) halpn = c(halpn, i)
      if(pns[i]>hpnthres) hahpn = c(hahpn, i)
    }
  }
  
  # Correlations and regression for low and high attitude agents -----------------------------------
  
  
  rbinL = NULL
  lR = NULL # regression among low attitude agents
  regL = NULL
  cl = NULL  # correlation among low attitude agents
  
  rbinH = NULL
  hR = NULL # regression among low attitude agents
  regH = NULL
  ch = NULL # correlation among low attitude agents
  
  # we have assumed that at least 5 agents should be in low and high attitude categories, otherwise this run will be ignored
  if((length(la) > 4) & (length(ha) > 4)){
    corrRegression(la, cl, regL, lR, rbinL, atts, sns, pns, ints)
    corrRegression(ha, ch, regH, hR, rbinH, atts, sns, pns, ints)
  }
  else{
    runCanBeUsed = 0
  }
  
  
  # representation of correlation and regression graph
  if(representation == 1 & runCanBeUsed == 1){
    chartCorrelation(lR)
    chartCorrelation(hR)
    chartRegression(lR[,3], lR[,4], rbinL)
    chartRegression(hR[,3], hR[,4], rbinH)    
  }
  
  return(c(mean(ints[lalpn]), mean(ints[lahpn]), mean(ints[halpn]), mean(ints[hahpn]), cl$r, cl$P, ch$r, ch$P, regL, regH, runCanBeUsed))
}

# compute  the linear regression model on perceived norm and intention. Also it calculates the correlation
# of attitudes, subjective norm, perceived norm and intention of agents
corrRegression = function(data, c, reg, aR, rbin, atts, sns, pns, ints){
  aTemp = giveTheDimensionTo(data, atts, sns, pns, ints) 
  cTemp = rcorr(as.matrix(aTemp))
  
  PNTemp = aTemp[,3]
  IntentionTemp = aTemp[,4]
  rbinTemp = hexbin(PNTemp, IntentionTemp)
  regTemp = (summary(lm(IntentionTemp~PNTemp)))$coefficients
  
  eval.parent(substitute(c <- cTemp))
  eval.parent(substitute(reg <- regTemp))
  eval.parent(substitute(aR <- aTemp))
  eval.parent(substitute(rbin <- rbinTemp))
}

# the inital attitudes of agents are cretaed by this function based on normal distribution
createAttitudes = function(nagents, sigma, delta){
  atts = NULL

  meanAtt = delta
  
  atts = NULL
  
  while(length(atts) < (nagents )) {
    atti = rnorm(1, meanAtt, sigma)
    if((atti < 1) & (atti > -1)){ 
      atts = c(atts, atti)
    }
  }
  return(atts)  
}

# representation of correlation in a plot
chartCorrelation = function(l){
  library(PerformanceAnalytics)
  chart.Correlation(as.matrix(l), histogram=TRUE, pch=19)
}

# representation of linear regression results in a plot
chartRegression = function(PN, Intention, rbin){
  library("Hmisc")
  library("hexbin")
  plot(PN, Intention)
  hexVP.abline(hexViewport(rbin),lm(Intention~PN), col="blue")  
}

# takes different vectors and store them in a matrix
giveTheDimensionTo = function(l, atts, sns, pns, ints){
  aR = c(atts[l], sns[l], pns[l], ints[l])
  dim(aR) <- c(length(aR)/4,4)
  colnames(aR) = c("Attitude","SN","PN","Intention")
  return(aR)
}

# each agent computes what it perceives from the attitude of others rather than what they are in reality

perceivedAttitude = function(myAttitude, att, th, rho){
  attP = 0
  if (th - abs(att - myAttitude) < 0) 
  {
    attP = (att - rho*(myAttitude - att)) 
  }
  else {
    attP = (att + rho*(myAttitude - att))
  }
  if(attP < -1)
    attP = -1
  else if (attP > 1)
    attP = 1
  return(attP)
}

pn_basedOn_threshold = function(rho, attss, th, myAttitude, nett2){
  sumPN = 0
  for(i in nett2)
  {sumPN = sumPN + (perceivedAttitude(myAttitude, attss[i], th, rho))
  }
  return(sumPN/(length(nett2)))
}

th = function(tauM, omega, intention, mean, beta){
  thh = tauM*(omega - beta * abs(intention - mean))
  return(thh)
}

# select subjective network based on probablity. That is those who hold more close attitude get more chance to be in the subjectie network
sn_prob = function(network, alpha, myAttitude, nSN, attss){
  proba = exp(-alpha * (myAttitude-attss[network])*(myAttitude-attss[network]))
  proba = proba/sum(proba)
  return(sample(network, size = nSN, replace = FALSE, prob = proba))
}

#--------------------------------------------------------------------------------------------------
#                abc indicator of data and simulation- accept reject fucntions
computeIndicator_RegFromData = function(){
  regsCoef = c(0.36225, -0.499) # from data (coef average of both datasets)
  return(regsCoef)
}

computeError_RegFromData = function(){
  regsErr = c(0.15195 * 2, 0.12438 * 2) # from data (sd average of both datasets)
  return(regsErr)
}

# the function evaluates the accepatnce of imput parameter set based on the distance to the average of bith datasets
# 
accept_basedOnReg = function(simRegRes){
  accept = FALSE
  reg_error_data = computeError_RegFromData()
  #print(reg_error_data)
  distanceSimData = distance_SimData(simRegRes)
  #print(distanceSimData)
  dL = abs(distanceSimData[1] - reg_error_data[1])
  dH = abs(distanceSimData[2] - reg_error_data[2])
  if(distanceSimData[1]<reg_error_data[1] & distanceSimData[2]<reg_error_data[2]){
    accept = TRUE 
  }
  return(accept)
}

distance_SimData = function(simRegRes){
  reg_coef_data = computeIndicator_RegFromData()
  
  d_regL = abs(simRegRes[1] - reg_coef_data[1])
  d_regH = abs(simRegRes[2] - reg_coef_data[2])
  
  d_simdata = c(d_regL, d_regH)
  return(d_simdata)
}
# -----------------------------------------------------------------------------------------

mismatchExploration = function(samples, nloops, nagents, n1_Range, n2_Range, methodSN_Range, allrest_Range, alpha_Range, coef_Range, rho_Range, beta_Range, tauM_Range, sigma_Range, delta_Range, representation, image_xy, steps){
  n1 = n1_Range[2]
  n1_Counter = 1
  n2 = n2_Range[2]
  n2_Counter = 1
  methodSN = methodSN_Range[1]
  methodSN_Counter = 1
  allrest = allrest_Range[1]
  allrest_Counter = 1
  alpha = alpha_Range[2]  
  alpha_Counter = 1
  coef = coef_Range[2]
  coef_Counter = 1
  rho = rho_Range[2]
  rho_Counter = 1
  beta = beta_Range[2]
  beta_counter = 1
  tauM = tauM_Range[2]
  tauM_Counter = 1
  sigma = sigma_Range[2]
  sigma_Counter = 1
  delta = delta_Range[2]
  delta_Counter = 1
  
  library(lhs)
  ABC <- randomLHS(samples, 6) 
  ABC[,1] <- round(qunif(ABC[,1], beta_Range[2], beta_Range[3]), digits = steps) # beta 
  ABC[,2] <- round(qunif(ABC[,2], tauM_Range[2], tauM_Range[3]), digits = steps) # taum
  ABC[,3] <- round(qunif(ABC[,3], sigma_Range[2], sigma_Range[3]), digits = steps) # sigma 
  ABC[,4] <- round(qunif(ABC[,4], delta_Range[2], delta_Range[3]), digits = steps) # delta
  ABC[,5] <- round(qunif(ABC[,5], rho_Range[2], rho_Range[3]), digits = steps) # rho
  ABC[,6] <- round(qunif(ABC[,6], alpha_Range[2], alpha_Range[3]), digits = steps) # alpha
  
  #1: number   4: input parametres   2:regl and regH  1:accpet or reject  
  sampleResults = array(0, dim = c(samples, 1+6+2+1))
  
  
  
  # --------------------------------------------------------------------------------
  # ------------- calling mismatchloop function for main computation----------------------------------
  x = 1
  while(x <= samples){
    res = c(0,0,0,0,NULL,NULL,NULL,NULL,NULL,NULL)
    res = mismatchloop(nloops, nagents, n1, n2, methodSN, allrest, ABC[x,6], coef, ABC[x,5], ABC[x,1], ABC[x,2], ABC[x,3], ABC[x,4]) 
    
    res = res/nloops
    clR = res[5:20]
    dim(clR) <- c(length(clR)/4,4)
    colnames(clR) = c("Attitude","SN","Intention","PN")
    
    clP = res[21:36]
    dim(clP) <- c(length(clP)/4,4)
    colnames(clP) = c("Attitude","SN","Intention","PN")
    
    chR = res[37:52]
    dim(chR) <- c(length(chR)/4,4)
    colnames(chR) = c("Attitude","SN","Intention","PN")
    
    chP = res[53:68]
    dim(chP) <- c(length(chP)/4,4)
    colnames(chP) = c("Attitude","SN","Intention","PN")
    
    regL = res[69:76]
    dim(regL) <- c(length(regL)/4,4)
    colnames(regL) = c("Estimate","std","tvalue","pvalue")
    
    regH = res[77:84]
    dim(regH) <- c(length(regH)/4,4)
    colnames(regH) = c("Estimate","std","tvalue","pvalue")
    
    sim_regResult = c(regL[2,1], regH[2,1])
    
    sampleResults[x, 1] = x
    sampleResults[x, 2] = ABC[x,1] #beta
    sampleResults[x, 3] = ABC[x,2] #taum   
    sampleResults[x, 4] = ABC[x,3] #sigma   
    sampleResults[x, 5] = ABC[x,4] #delta
    sampleResults[x, 6] = ABC[x,5] #rho
    sampleResults[x, 7] = ABC[x,6] #alpha
    sampleResults[x, 8] = regL[2,1] 
    sampleResults[x, 9] = regH[2,1]
    
    if(accept_basedOnReg(sim_regResult)){
      sampleResults[x, 10] = 1
    }else{
      sampleResults[x, 10] = 0
    }
    
    
    p = c("run ", x)
    if(x%%1000 == 0)
      print(p)
    x = x+1
  }
  return(sampleResults)
}


length_Range2 = function(c){
  len1 = ((c[3]-c[2])*1000)/(c[1]*1000)
  len = len1 + 1
  return(len)
}

length_Range = function(c){
  counter = 0
  v = c[2]
  while(v <= (c[3]+0.000001)){
    v = v + c[1]
    counter = counter + 1
  }
  #print(counter)
  return(counter)
}



#------------------------------------------------------------------------------------------
# for the matter of stochasticity, we run the main function mismatch for nloops times 
# and then making the average of all runs
mismatchloop = function(nloops, nagents, n1, n2, methodSN, allrest, alpha, coef,  rho, beta, tauM, sigma, delta){
  if((methodSN != "snP") & (methodSN != "snR") & (methodSN != "snN")){
    return("the 5th variable should be: snR, snP or snN")
  }
  if ((allrest != "all") & (allrest != "rest"))
    return("the 6th variable should be: all or rest")
  
  res = c(0,0,0,0,NULL,NULL,NULL,NULL,NULL,NULL)
  for (i in 1:nloops){
    res = res + mismatch(nloops, nagents, n1, n2, methodSN, allrest, "random", alpha, coef, rho, beta, tauM, sigma, delta) 
    #print(c("c: ", i))
  }
  
  return(res)
}

#------------------------------------------------------------------------------------------

# Run: to run the model, you need to compile all above functions, then run the mismatchExploration function below with the following inputs
# to run the model we need to initializae the input parameters

#the structure of each parameter is based on (steps, min, max). It includes the min and max too.
# for example c(0.2, 0.1, 0.5) will produce values {0.1, 0.3, 0.5}


# important point about c(steps, min, max): since we are using ABC, a random value is taken in the interval min and max
# and steps is not used.

n1_Range = c(1, 5, 5) # important others
n2_Range = c(50, 30, 30)
methodSN_Range = c("snP") # we aim to test the model based on other scenarios. snP stands for selecting important others based on probability
allrest_Range = c("all")  # The perceived group norm is computed based on all agents in the acquainted network of agent
alpha_Range = c(1, 0, 6)  
coef_Range = c(0.1, 0.2, 0.2)
rho_Range = c(0.1, 0, 1) #0.4

len = 2
stepsBeta = 0.999/len
stepsTaum = (2-1)/len
beta_Range = c(stepsBeta, 0.0, 0.999)
tauM_Range = c(stepsTaum, 1, 2)


sigma_Range = c(0.018, 0.2, 0.6)
delta_Range = c(0.015, 0.0, 0.75)

representation = 0 # for single run simulation, we set it to 1 for graph representation.
set.seed(3636)
samples = 25000000 # for running on cluster and faster execution, you can have different threads with smaller run samples
steps = 4 # the decimal number


resultsSamples = mismatchExploration(samples, 1, 1000, n1_Range, n2_Range, methodSN_Range, allrest_Range, alpha_Range, coef_Range, rho_Range, beta_Range, tauM_Range, sigma_Range, delta_Range, representation, len, steps)

write.csv(resultsSamples, "abc1.csv")



